/*
 * Copyright (c) 2019 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package apiclient

import (
	"fmt"
	"os"
	"time"

	"code.uplex.de/uplex-varnish/k8s-vcl-reloader/pkg/varnish"

	"github.com/sirupsen/logrus"

	"k8s.io/apimachinery/pkg/api/meta"
	"k8s.io/apimachinery/pkg/runtime"
	utilruntime "k8s.io/apimachinery/pkg/util/runtime"
	"k8s.io/apimachinery/pkg/util/wait"
	"k8s.io/client-go/informers"
	"k8s.io/client-go/kubernetes"
	core_v1 "k8s.io/client-go/kubernetes/typed/core/v1"
	core_v1_listers "k8s.io/client-go/listers/core/v1"
	"k8s.io/client-go/tools/cache"
	"k8s.io/client-go/tools/record"

	api_v1 "k8s.io/api/core/v1"
	meta_v1 "k8s.io/apimachinery/pkg/apis/meta/v1"
)

// ReloaderCfg represents the user configuration for VCL reloading.
type ReloaderCfg struct {
	Templates map[string]string
	Services  []string
	CfgMap    string
	Main      string
}

type infrmrs struct {
	svc    cache.SharedIndexInformer
	endp   cache.SharedIndexInformer
	cfgMap cache.SharedIndexInformer
}

// SyncType classifies the sync event, passed through to workers.
type SyncType uint8

const (
	// Add event
	Add SyncType = iota
	// Update event
	Update
	// Delete event
	Delete
)

// SyncObj wraps the object for which event handlers are notified, and
// encodes the sync event. These are the objects passed into the
// queues for workers.
type SyncObj struct {
	Type SyncType
	Obj  interface{}
}

// Listers aggregates listers from k8s.io/client-go/listers for the
// various resource types of interest. These are initialized by
// Client, and handed off to NamespaceWorker workers to read data from
// the client-go cache.
type Listers struct {
	svc    core_v1_listers.ServiceLister
	endp   core_v1_listers.EndpointsLister
	cfgMap core_v1_listers.ConfigMapLister
}

// Client watches Kubernetes API and reconfigures Varnish via
// varnish.Controller when needed.
type Client struct {
	log       *logrus.Logger
	client    kubernetes.Interface
	informers *infrmrs
	listers   *Listers
	pod       *api_v1.Pod
	nsQs      *NamespaceQueues
	vc        *varnish.Controller
	stopCh    chan struct{}
	recorder  record.EventRecorder
	svcs      map[string]struct{}
	cfgMap    string
	podNs     string
	podName   string
}

// NewClient creates an API client.
//
//    log: logger initialized at startup
//    kubeClient: k8s client initialized at startup
//    vc: Varnish controller
//    infFactory: SharedInformerFactory to create informers & listers for
//                the k8s standard client APIs
func NewClient(
	log *logrus.Logger,
	kubeClient kubernetes.Interface,
	updaterCfg SpecUpdaterConfig,
	vc *varnish.Controller,
	infFactory informers.SharedInformerFactory,
	svcs []string,
	cfgMap string,
	podNs string,
	podName string) (*Client, error) {

	apiclient := Client{
		log:     log,
		client:  kubeClient,
		vc:      vc,
		stopCh:  make(chan struct{}),
		svcs:    make(map[string]struct{}),
		cfgMap:  cfgMap,
		podNs:   podNs,
		podName: podName,
	}

	for _, svc := range svcs {
		apiclient.svcs[svc] = struct{}{}
	}

	InitMetrics()

	eventBroadcaster := record.NewBroadcaster()
	eventBroadcaster.StartLogging(apiclient.log.Printf)
	eventBroadcaster.StartRecordingToSink(&core_v1.EventSinkImpl{
		Interface: apiclient.client.CoreV1().Events(""),
	})
	evtScheme := runtime.NewScheme()
	if err := api_v1.AddToScheme(evtScheme); err != nil {
		return nil, err
	}
	apiclient.recorder = eventBroadcaster.NewRecorder(evtScheme,
		api_v1.EventSource{Component: "vcl-reloader"})

	apiclient.informers = &infrmrs{
		svc:    infFactory.Core().V1().Services().Informer(),
		endp:   infFactory.Core().V1().Endpoints().Informer(),
		cfgMap: infFactory.Core().V1().ConfigMaps().Informer(),
	}

	evtFuncs := cache.ResourceEventHandlerFuncs{
		AddFunc:    apiclient.addObj,
		DeleteFunc: apiclient.deleteObj,
		UpdateFunc: apiclient.updateObj,
	}

	apiclient.informers.svc.AddEventHandler(evtFuncs)
	apiclient.informers.endp.AddEventHandler(evtFuncs)
	apiclient.informers.cfgMap.AddEventHandler(evtFuncs)

	apiclient.listers = &Listers{
		svc:    infFactory.Core().V1().Services().Lister(),
		endp:   infFactory.Core().V1().Endpoints().Lister(),
		cfgMap: infFactory.Core().V1().ConfigMaps().Lister(),
	}

	updater := NewSpecUpdater(log, updaterCfg, vc, apiclient.listers)

	apiclient.nsQs = NewNamespaceQueues(apiclient.log, updater,
		apiclient.listers, apiclient.recorder)

	return &apiclient, nil
}

func (apiclient *Client) logObj(action string, obj interface{}) {
	apiclient.log.Debug(action, ":", obj)
	m, mErr := meta.Accessor(obj)
	t, tErr := meta.TypeAccessor(obj)
	if mErr == nil && tErr == nil {
		if t.GetKind() != "" {
			apiclient.log.Debugf("%s %s: %s/%s", action,
				t.GetKind(), m.GetNamespace(), m.GetName())
		} else {
			apiclient.log.Debugf("%s: %s/%s", action,
				m.GetNamespace(), m.GetName())
		}
	}
}

func (apiclient *Client) ignoring(obj interface{}) bool {
	m, err := meta.Accessor(obj)
	if err != nil {
		apiclient.log.Errorf("Cannot read metadata for sync object: %s",
			err)
		return false
	}
	key := fmt.Sprintf("%s/%s", m.GetNamespace(), m.GetName())

	switch obj.(type) {
	case *api_v1.Service, *api_v1.Endpoints:
		_, exists := apiclient.svcs[key]
		return !exists
	case *api_v1.ConfigMap:
		return key != apiclient.cfgMap
	}
	return false
}

func (apiclient *Client) logIgnore(obj interface{}) {
	ns := "Unknown"
	name := "Unknown"
	kind := "Unknown"
	m, mErr := meta.Accessor(obj)
	t, tErr := meta.TypeAccessor(obj)
	if mErr != nil {
		apiclient.log.Errorf("Cannot read metadata for sync object: %s",
			mErr)
	} else {
		ns = m.GetNamespace()
		name = m.GetName()
	}
	if tErr != nil {
		apiclient.log.Errorf("Cannot read type for sync object: %s",
			tErr)
	} else if t.GetKind() != "" {
		kind = t.GetKind()
	} else {
		switch obj.(type) {
		case *api_v1.Service:
			kind = "Service"
		case *api_v1.Endpoints:
			kind = "Endpoints"
		case *api_v1.ConfigMap:
			kind = "ConfigMap"
		}
	}
	apiclient.log.Debugf("Ignoring %s %s/%s", kind, ns, name)
	syncCounters.WithLabelValues(ns, kind, "Ignore").Inc()
}

func incWatchCounter(sync string, obj interface{}) {
	switch obj.(type) {
	case *api_v1.Service:
		watchCounters.WithLabelValues("Service", sync).Inc()
	case *api_v1.Endpoints:
		watchCounters.WithLabelValues("Endpoints", sync).Inc()
	case *api_v1.ConfigMap:
		watchCounters.WithLabelValues("ConfigMap", sync).Inc()
	default:
		watchCounters.WithLabelValues("Unknown", sync).Inc()
	}
}

func (apiclient *Client) addObj(obj interface{}) {
	apiclient.logObj("Add", obj)
	incWatchCounter("Add", obj)
	if apiclient.ignoring(obj) {
		apiclient.logIgnore(obj)
		return
	}
	apiclient.nsQs.Queue.Add(&SyncObj{Type: Add, Obj: obj})
}

func (apiclient *Client) deleteObj(obj interface{}) {
	apiclient.logObj("Delete", obj)
	incWatchCounter("Delete", obj)
	if apiclient.ignoring(obj) {
		apiclient.logIgnore(obj)
		return
	}
	apiclient.nsQs.Queue.Add(&SyncObj{Type: Delete, Obj: obj})
}

func (apiclient *Client) updateObj(old, new interface{}) {
	apiclient.log.Debug("Update:", old, new)
	incWatchCounter("Update", new)
	if apiclient.ignoring(new) {
		apiclient.logIgnore(new)
		return
	}
	oldMeta, oldErr := meta.Accessor(old)
	newMeta, newErr := meta.Accessor(new)
	t, tErr := meta.TypeAccessor(old)
	if oldErr == nil && newErr == nil &&
		oldMeta.GetResourceVersion() == newMeta.GetResourceVersion() {
		apiclient.log.Debugf("Update %s/%s: unchanged",
			oldMeta.GetNamespace(), oldMeta.GetName())
		apiclient.logIgnore(new)
		return
	}

	// kube-system resources frequently update Endpoints with
	// empty Subsets, ignore them.
	oldEndp, oldEndpExists := old.(*api_v1.Endpoints)
	newEndp, newEndpExists := new.(*api_v1.Endpoints)
	if oldEndpExists && newEndpExists &&
		len(oldEndp.Subsets) == 0 && len(newEndp.Subsets) == 0 {

		apiclient.log.Debugf("Update endpoints %s/%s: empty Subsets, "+
			"ignoring", newEndp.Namespace, newEndp.Name)
		syncCounters.WithLabelValues(oldMeta.GetNamespace(),
			"Endpoints", "Ignore").Inc()
		return
	}

	var metaObj *meta_v1.Object
	if oldErr == nil {
		metaObj = &oldMeta
	} else if newErr == nil {
		metaObj = &newMeta
	}
	if metaObj != nil {
		if tErr == nil && t.GetKind() != "" {
			apiclient.log.Debugf("Update %s: %s/%s", t.GetKind(),
				(*metaObj).GetNamespace(), (*metaObj).GetName())
		} else {
			apiclient.log.Debugf("Update: %s/%s",
				(*metaObj).GetNamespace(), (*metaObj).GetName())
		}
	}
	apiclient.nsQs.Queue.Add(&SyncObj{Type: Update, Obj: new})
}

// Run the API client -- start the informers in goroutines, wait for
// the caches to sync, and call Run() for the NamespaceQueues. Then
// block until Stop() is invoked.
//
// If readyFile is non-empty, it is the path of a file to touch when
// the controller is ready (after informers have launched).
func (apiclient *Client) Run(readyFile string, metricsPort uint16) {
	defer utilruntime.HandleCrash()
	defer apiclient.nsQs.Stop()

	apiclient.log.Info("Launching informers")
	go apiclient.informers.svc.Run(apiclient.stopCh)
	go apiclient.informers.endp.Run(apiclient.stopCh)
	go apiclient.informers.cfgMap.Run(apiclient.stopCh)

	apiclient.log.Infof("Starting metrics listener at port %d", metricsPort)
	go ServeMetrics(apiclient.log, metricsPort)

	apiclient.log.Info("k8s API client ready")
	if readyFile != "" {
		f, err := os.Create(readyFile)
		if err != nil {
			e := fmt.Errorf("Cannot create ready file %s: %v",
				readyFile, err)
			utilruntime.HandleError(e)
			return
		}
		if err = f.Close(); err != nil {
			e := fmt.Errorf("Cannot close ready file %s: %v",
				readyFile, err)
			utilruntime.HandleError(e)
			defer f.Close()
		}
		apiclient.log.Infof("Created ready file %s", readyFile)
	}

	apiclient.log.Info("Waiting for caches to sync")
	if ok := cache.WaitForCacheSync(apiclient.stopCh,
		apiclient.informers.svc.HasSynced,
		apiclient.informers.endp.HasSynced,
		apiclient.informers.cfgMap.HasSynced); !ok {

		err := fmt.Errorf("Failed waiting for caches to sync")
		utilruntime.HandleError(err)
		return
	}

	apiclient.log.Info("Caches synced, running workers")
	apiclient.log.Info("Starting initial update")
	apiclient.vc.SetUpdater(apiclient.nsQs.updater)
	if err := apiclient.nsQs.updater.Update(); err != nil {
		apiclient.log.Errorf("Initial update failed: %s", err)
	}
	go wait.Until(apiclient.nsQs.Run, time.Second, apiclient.stopCh)

	<-apiclient.stopCh
}

// Stop the VCL controller -- signal the workers to stop.
func (apiclient *Client) Stop() {
	apiclient.stopCh <- struct{}{}
	apiclient.log.Info("Shutting down workers")
	close(apiclient.stopCh)
	<-apiclient.nsQs.DoneChan
	apiclient.log.Info("k8s API client exiting")
}
